// USB-DMX-G431  ***************************************************************
//
// Simple DMX512 controller for STM32G431KB (32 pin LQFP package)
// based on USB MIDISTREAM and CDC-ABSTRACT CONTROL MODEL
//
// supports the MiniDMX protocol 
// plus simple MIDI and ASCII string based command sets
//
// Author and (c)2020-21 Wolfgang Schemmert. 
// The copyright claim and GPL-3 licence is restricted to this file
// and the depending files "main.h","usb-fs-431.c","dmx.c","userinterface.c"
// and "globaldefines.h".
// Other depending material is published and copyrighted
// by STMicroelectronics and Segger GmbH/Rowley Associates Ltd
//
// Contact: <www.midi-and-more.de>
// A detailled hardware construction 
// and operation manual is provided at this website
//
// ****************************************************************************
//      This program is free software: you can redistribute it and/or modify
//      it under the terms of the GNU General Public License as published by
//      the Free Software Foundation, either version 3 of the License, or
//      (at your option) any later version.
//
//      This program is distributed in the hope that it will be useful,
//      but WITHOUT ANY WARRANTY; without even the implied warranty of
//      MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//      GNU General Public License for more details.
//
//      You should have received a copy of the GNU General Public License
//      along with this program.  If not, see <http://www.gnu.org/licenses/>.
//
//      Licence: GPL-3
//
// ****************************************************************************

#include "stm32g431xx.h"
#include "globaldefines.h"
#include "main.h"

int main(void)    // ********************************************************
// **************************************************************************
{
//uint32_t k;
uint32_t recvword;
uint8_t recvbyte, index, status, data1, data2;

	RCC->APB1ENR1 |= RCC_APB1ENR1_PWREN;  //necessary for what??
	RCC->AHB2ENR |= RCC_AHB2ENR_GPIOAEN;  // Enable Port A clock
	RCC->AHB2ENR |= RCC_AHB2ENR_GPIOBEN;  // Enable Port B clock

	GPIOA->MODER = 0x28000000;      //all input except PA14,13
	GPIOB->MODER = 0x00000000;      //primarily all input
	GPIOA->PUPDR = 0x9601AAAA;      //primarily all inputs pulldown,
									//except PA13(pullUp) and PA9,10,11,12,13,14
	GPIOB->PUPDR = 0xAAAAAAAA;      //primarily all inputs pulldown
	SetSysClock();

	gSYSCOUNT = 20;
	gLEDCOUNT = 0;
	gDELAYCOUNT = 0;
	gTIMEOUT = 0;
    gLedBlink = 0;
	gVid = DEFAULT_VID;
	gPid = DEFAULT_PID;
    guDeviceClass = CDC_CLASS;
	gChannel = 0;
	gdFadeTime = 0;
	gdDmxMax = 512;
    gdMiniDmxOn = 0;
	gdGlobalSat = 100;
	gdGlobalLum = 100;

	SysTick_Config(SystemCoreClock/1000);

//start Independent Watchdog  **************
	IWDG->KR = 0x00005555;
	IWDG->RLR = 0x0000FFFF;
	msDelay(2);
	IWDG->PR = 0x00000002;		//resulting timeout ca. 1-2 seconds
	msDelay(2);
	IWDG->KR = 0x0000CCCC;		//start Independent Watchdog

	msDelay(500);
	gTest = 0;
	if ((GPIOA->IDR & 0x00002000) == 0)  {	//PA13 shorted to Ground
//then reset global parameters to default
		pushFlashPage8(0xA5E7B3C9);			//erase only
		msDelay(500);
		gTest = 1;
	}

//re-configure I/O pins specifically
//Configure PA8 as LED driver
	GPIOA->MODER &= ~(GPIO_MODER_MODE8);
	GPIOA->MODER |= GPIO_MODER_MODE8_0 ;  // Set PortA8 (LED) as output
	GPIOA->BSRR =  GPIO_BSRR_BS_8;        // PA8 (LED) ON

//check jumper or front panel switch and decide if ASCII or MIDI commands ****
    GPIOA->MODER &= ~(GPIO_MODER_MODE14);	//repeat input mode
    GPIOA->PUPDR |=  GPIO_PUPDR_PUPDR14_0;	//does not conflict with SWDebug
    GPIOA->PUPDR &=  ~GPIO_PUPDR_PUPDR14_1;
    msDelay(10);
//if PortA,14 = high (default): CDC_CLASS
//if PortA,14 = low (Jumper set): MIDI_CLASS
    guDeviceClass = CDC_CLASS;
    if ((GPIOA->IDR & GPIO_IDR_ID14) == 0) {
        guDeviceClass = MIDI_CLASS;
    }
#ifdef RS232_DEBUG
    initUsart2();
    msDelay(20);
    sendDBGB('@');
#endif

//init I/O state  *********************
	popFlashPage8();           //load only global system parameters
//init communication  *********************
#ifndef DMX_DEBUG
// init DMX transmitter
	popDmxFlash(0);
	msDelay(10);
	startDmxTx();
	DmxHeader();
	msDelay(10);
#endif
	guConfigured = 0;
	initUSB();
	gTIMEOUT = 5000;
	while (gTIMEOUT)  {
		servSystem();   //serve DMX process, LED, watchdog, USB mode
		if (guConfigured)  break;
	}
	msDelay(10);

    gcPushByte = 0;

	while(1)
    {
        servSystem();                       //essential, controls DMX timing
        if (guDeviceClass == CDC_CLASS)  {
            if (gcPushByte)  {
                recvbyte = gcPushByte;      //uses next command code as
                gcPushByte = 0;             //flexible parameter input delimiter
				ASCII_Interpreter(recvbyte);
            }
            recvword = pullUsbRxBuf();
            if (recvword)  {
                recvbyte = (uint8_t)(recvword & 0x00FF);
                if (recvbyte == 'Z')  {
                    gTIMEOUT = 1000;
                    gdFaderBusy = 0;        //block fade process
                    MiniDMX_Interpreter();
                }
				else  ASCII_Interpreter(recvbyte);
            }
            else if ((gTIMEOUT == 0) && gdMiniDmxOn) {
                gdMiniDmxOn = 0;            //terminate MiniDMX
                gLedBlink = 0;
            }
        }
        else {
            recvword = pullUsbRxBuf();
            if (recvword)  {
                status = (uint8_t)(recvword & 0x00FF);
#ifdef DMX_DEBUG
				sendDBGB('s');
				sendX8(status);
#endif
                index = (status >> 4) & 0x0000000F;
                if (index > 0x0C) index = 0;

                recvword = pullUsbRxBuf();
                if (recvword)  {
					data1 = (uint8_t)(recvword & 0x00FF);
#ifdef DMX_DEBUG
					sendX8(data1);
#endif
                }
                else index = 0;
				if ((index != 0x0C) && (index != CRR)) {
					recvword = pullUsbRxBuf();
					if (recvword)  {
						data2 = (uint8_t)(recvword & 0x00FF);
#ifdef DMX_DEBUG
						sendX8(data2);
#endif
					}
					else  index = 0;
				}
				else data2 = 0;	//formally
#ifdef DMX_DEBUG
						sendDBGB('i');
						sendX8(index);
#endif

                if (index) MIDI_Interpreter(index,status,data1,data2);
            }
        }
    }
}

//##########################################################################

// Misc Low Level Routines
//************************

void SetSysClock(void)  // ***************************************************
//assumes "Vcore range 1", HSE clock 8.000 MHz, SystemClock=AHB=48MHz
//APB1+2 = 48MHz
//adopted from "system_stm32f0xx.c" and modified according to STM32G431 FRefMan
//Reset SW[1:0], HPRE[3:0], PPRE1[2:0], PPRE2[2:0], MCOSEL[2:0] and MCOPRE[2:0]
// **************************************************************************
{
uint32_t StartUpCounter = 0, OscStatus = 0;

#define STARTUP_TIMEOUT   0x50000 //Time out for HSE start up

    RCC->APB1ENR1 |= RCC_APB1ENR1_PWREN;  //Power Interface clock enable
//Vcore Range1 is default after Reset

// SYSCLK, HCLK, PCLK configuration ----------------------------------------
// Enable HSE

    RCC->CR |= RCC_CR_HSEON;

// Wait till HSE is ready or Timeout is reached
    do  {
      OscStatus = RCC->CR & RCC_CR_HSERDY;
      StartUpCounter++;
     } while((OscStatus == 0) && (StartUpCounter  < STARTUP_TIMEOUT));

#ifdef CPU_CLOCK144
    if (OscStatus) {
// Enable Prefetch Buffer and set Flash Latency
	FLASH->ACR =
     FLASH_ACR_PRFTEN| FLASH_ACR_DCEN |FLASH_ACR_ICEN |FLASH_ACR_LATENCY_7WS;

//System Clock shall be 144MHz
//  PLL_VCO = (HSE_VALUE or HSI_VALUE / PLL_M) * PLL_N
#define PLL_M      0        //8 MHz PLL input for project
#define PLL_N      36       //8*36=288MHz for project
// USB OTG FS, SDIO and RNG Clock =  PLL_VCO / PLLQ
#define PLL_Q      2        //divides PLL/6 for 48MHz USB clock
// SYSCLK = PLL_VCO / PLL_P
#define PLL_R      0        //divides PLL/2 for 144MHz SYSCLK

 // HCLK = SYSCLK/
    RCC->CFGR |= RCC_CFGR_HPRE_DIV1;    //set for project to 144MHz
 // PCLK2 = HCLK / 2
    RCC->CFGR |= RCC_CFGR_PPRE2_DIV2;   //set for project to 72MHz
 //PCLK1 = HCLK / 2
    RCC->CFGR |= RCC_CFGR_PPRE1_DIV2;   //set for project to 72MHz

//Configure the main PLL
//PLL input is 8MHz, so PLL_M is set correctly with next command:
//implies correct setting of PLL_R and PLL_Q, too
     RCC->PLLCFGR = PLL_N << 8;         //PLL_N = 36=0x12 = 001 0010 (bits 8-14)
     RCC->PLLCFGR |= PLL_Q << 21;        //PLL_Q = 2 (bits 21,22)
//    RCC->PLLCFGR |= PLL_R << 25;        //PLL_R = 00  (bits 25,26)

    RCC->CCIPR &= ~RCC_CCIPR_CLK48SEL;
    RCC->CCIPR |= RCC_CCIPR_CLK48SEL_1;            //select PLL_Q as USB clock
    RCC->PLLCFGR |= (RCC_PLLCFGR_PLLQEN | RCC_PLLCFGR_PLLREN);

#else  //CPU_CLOCK = 48MHz
    if (OscStatus) {
// Enable Prefetch Buffer and set Flash Latency
	FLASH->ACR =
        FLASH_ACR_PRFTEN| FLASH_ACR_DCEN |FLASH_ACR_ICEN |FLASH_ACR_LATENCY_2WS;

//System Clock shall be 48MHz
//  PLL_VCO = (HSE_VALUE or HSI_VALUE / PLL_M) * PLL_N
#define PLL_M      0        //8 MHz PLL input for project
#define PLL_N      12       //8*12=96MHz for project
// USB OTG FS, SDIO and RNG Clock =  PLL_VCO / PLLQ
#define PLL_Q      0        //divides PLL/2 for 48MHz USB clock
// SYSCLK = PLL_VCO / PLL_P
#define PLL_R      0        //divides PLL/2 for 48MHz SYSCLK

 // HCLK = SYSCLK/
	RCC->CFGR |= RCC_CFGR_HPRE_DIV1;    //set for project to 48MHz
 // PCLK2 = HCLK / 2
	RCC->CFGR |= RCC_CFGR_PPRE2_DIV1;   //set for project to 48MHz
 //PCLK1 = HCLK / 2
	RCC->CFGR |= RCC_CFGR_PPRE1_DIV1;   //set for project to 48MHz

//Configure the main PLL
//PLL input is 8MHz, so PLL_M is set correctly with next command:
//implies correct setting of PLL_R and PLL_Q, too
	RCC->PLLCFGR = PLL_N << 8;	//PLL_N = 12=0x0C = 000 1100 (bits 8-14)
//    RCC->PLLCFGR |= PLL_Q << 21;	//PLL_Q = 00 (bits 21,22)
//    RCC->PLLCFGR |= PLL_R << 25;	//PLL_R = 00  (bits 25,26)

	RCC->CCIPR &= ~RCC_CCIPR_CLK48SEL;
	RCC->CCIPR |= RCC_CCIPR_CLK48SEL_1; //select PLL_Q as USB clock
	RCC->PLLCFGR |= (RCC_PLLCFGR_PLLQEN |  RCC_PLLCFGR_PLLREN);
#endif
//Select HSE as PLL source
	RCC->PLLCFGR |= RCC_PLLCFGR_PLLSRC_HSE; //set bits 0,1

// Enable the main PLL
	RCC->CR |= RCC_CR_PLLON;
// Wait till the main PLL is ready
	while((RCC->CR & RCC_CR_PLLRDY) == 0) ;

// Select PLL as system clock source
     RCC->CFGR |= RCC_CFGR_SW_PLL;

// Wait till PLL is used as system clock source
	while ((RCC->CFGR & RCC_CFGR_SWS) != RCC_CFGR_SWS_PLL) ;

	SystemCoreClockUpdate(); //update in correspondence with parameters set above

	RCC->AHB1ENR |= RCC_AHB1ENR_FLASHEN; //necessary for Flash read/write action
    }


#ifndef  CPU_CLOCK144
    else  {
//init HSI, synchronize with USB, if possible
		RCC->CR &= ~RCC_CR_HSEON;
		RCC->CR |= RCC_CR_HSION;

// Wait till HSI is ready or Timeout is reached
		do  {
		OscStatus = RCC->CR & RCC_CR_HSIRDY;
		StartUpCounter++;
		} while((OscStatus == 0) && (StartUpCounter  < STARTUP_TIMEOUT));

		if (OscStatus) {
// Enable Prefetch Buffer and set Flash Latency
			FLASH->ACR = FLASH_ACR_PRFTEN| FLASH_ACR_DCEN |
										FLASH_ACR_ICEN |FLASH_ACR_LATENCY_2WS;

//System Clock shall be 96MHz
// USB OTG FS, SDIO and RNG Clock =  PLL_VCO / PLLQ
#define PLL_MI      1        //8 MHz PLL input for project
#define PLL_NI      12       //8*12=96MHz for project
// USB OTG FS, SDIO and RNG Clock =  PLL_VCO / PLLQ
#define PLL_QI      0        //divides PLL/2 for 48MHz USB clock
// SYSCLK = PLL_VCO / PLL_P
#define PLL_RI      0        //divides PLL/2 for 48MHz SYSCLK

 // HCLK = SYSCLK/
	RCC->CFGR |= RCC_CFGR_HPRE_DIV1;    //set for project to 48MHz
 // PCLK2 = HCLK / 2
	RCC->CFGR |= RCC_CFGR_PPRE2_DIV1;   //set for project to 48MHz
 //PCLK1 = HCLK / 2
	RCC->CFGR |= RCC_CFGR_PPRE1_DIV1;   //set for project to 48MHz

//Configure the main PLL
			RCC->PLLCFGR = PLL_MI << 4;
			RCC->PLLCFGR = PLL_NI << 8;		//PLL_N = 36=0x12 (bits 8-14)
			RCC->PLLCFGR |= PLL_QI << 21;	//PLL_Q = 2 (bits 21,22)
			RCC->PLLCFGR |= PLL_RI << 25;	//PLL_R = 00  (bits 25,26)

			RCC->CCIPR &= ~RCC_CCIPR_CLK48SEL;
			RCC->CCIPR |= RCC_CCIPR_CLK48SEL_1; //select PLL_Q as USB clock
			RCC->PLLCFGR |= (RCC_PLLCFGR_PLLQEN |  RCC_PLLCFGR_PLLREN);

//Select HSI as PLL source
			RCC->PLLCFGR |= RCC_PLLCFGR_PLLSRC_HSI; //set bit 1 only
// Enable the main PLL
			RCC->CR |= RCC_CR_PLLON;
// Wait till the main PLL is ready
			while((RCC->CR & RCC_CR_PLLRDY) == 0) ;
// Select PLL as system clock source
			RCC->CFGR |= RCC_CFGR_SW_PLL;

// Wait till PLL is used as system clock source
			while ((RCC->CFGR & RCC_CFGR_SWS) != RCC_CFGR_SWS_PLL) ;

			SystemCoreClockUpdate();
							//update in correspondence with parameters set above

			RCC->AHB1ENR |= RCC_AHB1ENR_FLASHEN;
							//necessary for Flash read/write action

		}
    }
#endif
}


void LED_Off(void)  // ******************************************************
// built as subroutine to supply service to "usb-fs-431.c" hardware independent
// **************************************************************************
{
        GPIOA->BSRR = GPIO_BSRR_BR_8; // clear bit PA8,LED OFF
        gLEDCOUNT = 200; //milliseconds
}


void msDelay(uint32_t count)   // *******************************************
//blocking delay by 'count' milliseconds
// **************************************************************************
{
	gDELAYCOUNT = count;
	while(gDELAYCOUNT)  {
		IWDG->KR = 0x0000AAAA;  //timeout ca. 1-2 seconds
		servSystem();
	}
}


void wait5us(void)   // *****************************************************
// blocking delay ca 5us
// *************************************************************************
{
uint32_t countdown;

#ifdef CPU_CLOCK144
	countdown = 87;
#else
	countdown = 51;
#endif

    while (countdown)   {
        IWDG->KR = 0x0000AAAA;  //timeout ca. 1-2 seconds
        countdown-- ;
    }
}


void servSystem(void)   // ************************************************
//system relevant, but not extremely time critical services
//are handled quasi in the background
//especially called in pullUsbRxBuf
//but watchdog resets when it hangs somewhere else
//additionally slowed down by SysTick to 50 times/second
// ****************************************************************************
{
uint8_t classChange;
static uint8_t secCount ,deciCount;

//max. high speed part:
#ifndef DMX_DEBUG
	if (USART1->ISR & USART_ISR_TXE) {
		USART1->TDR = gdDMXBUF[gdDmxTxIndex];
// the TXE bit is cleared automatically when TDR is written
		gdDmxTxIndex++;
		if (gdDmxTxIndex >= gdDmxMax)  DmxHeader() ;
	}
#endif
//lower speed services:  **********************************

	if (gSYSCOUNT == 0)  {		//countdown in SysTick_Handler()
		gSYSCOUNT = 20;			//restart 50Hz service interval

		IWDG->KR = 0x0000AAAA;  //fed watchdog - timeout ca. 1-2 seconds

//handle DMX faders
		if (gdFaderBusy)  fadeProcess();

//LED control
		if (gLedBlink)  {
			if (gLedBlink < 8 ) GPIOA->BSRR = GPIO_BSRR_BS_8;
			else GPIOA->BSRR = GPIO_BSRR_BR_8;
			gLedBlink++ ;
			if (gLedBlink >= 12) gLedBlink = 1;
		}
		else if (gLEDCOUNT==0)  GPIOA->BSRR = GPIO_BSRR_BS_8;
													// Set bit PA8 (LED ON)

//10Hz service for Chaser   *******************
        deciCount ++;
        if (deciCount >= 5)  {
            deciCount = 0;

//			sendUsbX8((uint8_t)gdChaseTime);

			if (gdChaseTime && gdChaseLen)  {
				gdChaseCount--;
				if ((gdChaseCount == 0) ||		//>=0x100 = underflow rescue
								(gdChaseCount >= 0x100)) nextChaserStep();
			}

//very low speed service (1 per sec)    *******************
			secCount ++;
			if (secCount >= 10)  {
				secCount = 0;

				if (gTest) sendUsbByte('R');		//message reset of Globals

#ifndef DMX_DEBUG
	//check panel switch or jumper and force new USB enumeration if necessary
	//switch open: = (default) ASCII commands, Mini DMX via USB
	//switch closed: =  MIDI commands via USB
				classChange = 0;
				if (GPIOA->IDR & GPIO_IDR_ID14)  {        //then switch is open
					if (guDeviceClass == MIDI_CLASS) {
						stopUSB();
						guDeviceClass = CDC_CLASS;
						classChange = 0xFF;
					}
				}
				else if (guDeviceClass == CDC_CLASS) {
					stopUSB();
					guDeviceClass = MIDI_CLASS;
					classChange = 0xFF;
				}
				if (classChange)  {
					initUSB();
					gTIMEOUT = 5000;                    //5 seconds
					while (gTIMEOUT)  {
						IWDG->KR = 0x0000AAAA;          //feed watchdog
						if (guConfigured)  break;
					}
				}
#endif
			}
		}
    }
}


void popFlashPage8(void)   // **********************************************
//read project independent system parameters
// **************************************************************************
{
uint32_t * flashPtr;
uint32_t test, vid, pid;
//uint32_t *sramPtr;

	flashPtr = (uint32_t *)PAGE8;

	test = ~(*flashPtr);//virgin test cannot be==0 if project was saved in Flash
	flashPtr++;
//	vid =  ~(*flashPtr);    //cannot be==0 if any project was saved in Flash
	flashPtr++;
//	pid =  ~(*flashPtr);    //cannot be==0 if any project was saved in Flash
	flashPtr++;
//removed due to stability problems
//	if ((test == 0) | (vid == 0) | (pid == 0)) return;
								//Flash is virgin, load default empty project


	gChannel =  ~(*flashPtr);
	flashPtr++;
	gdDmxMax = ~*flashPtr;
	flashPtr++;
	gdFadeTime = ~*flashPtr;
	flashPtr++;
	vid=  ~(*flashPtr);
	flashPtr++;
	pid = ~*flashPtr;
	flashPtr++;

	if ((vid != 0) && (pid != 0))  {
		gVid = vid;
		gPid = pid;
	}
	else  {
		gVid = DEFAULT_VID;
		gPid = DEFAULT_PID;
	}
	if (gChannel > 15) gChannel = 15;
	if (gdDmxMax == 0) gdDmxMax = 512;			//virgin flash;
	if (gdDmxMax < 24) gdDmxMax = 24;
	if (gdDmxMax > 512) gdDmxMax = 512;
	if (gdFadeTime > 635) gdFadeTime = 635;		//127*5
}


uint32_t pushFlashPage8(uint32_t mode)   //**********************************
//save project independent system parameters permanently
// **************************************************************************
{
uint32_t *flashPtr;

//unlock flash
#define KEY1    0x45670123      //mandatory unlock key defined by STM
#define KEY2    0xCDEF89AB

    while ((FLASH->SR &FLASH_SR_BSY) !=0); //wait until no operation is going on
    if ((FLASH->CR & FLASH_CR_LOCK) != 0) {
		FLASH->KEYR = KEY1;
		FLASH->KEYR = KEY2;
    }
    else return 0;

//feed watchdog
    IWDG->KR = 0x0000AAAA;          //timeout ca. 1-2 seconds

//erase Page8
    while ((FLASH->SR & FLASH_SR_BSY) != 0) ;  //wait until flash is not busy
    FLASH->CR = ((8<<3) | FLASH_CR_EOPIE | FLASH_CR_ERRIE | FLASH_CR_PER);
                                //EOPIE is set formally to get SR_EOP set
    FLASH->CR |= FLASH_CR_STRT;
    while ((FLASH->SR & FLASH_SR_BSY) != 0) ;
    FLASH->SR |= FLASH_SR_EOP;
    if (FLASH->SR & FLASH_SR_OPERR)  return 0;

    FLASH->CR &= ~FLASH_CR_PER;
	msDelay(20);
	if (mode == 0xA5E7B3C9)  {  //erase only requested
//lock flash
		FLASH->CR |= FLASH_CR_LOCK;
		msDelay(100);
		return  1;		
	}

//write mirror into Page8:
//this microcontroller allows exclusively doubleword-wide write in flash
//only write USB Vid/Pid, gChannel, gdDxmMax into page7
 	FLASH->CR |= FLASH_CR_PG;
    flashPtr = (uint32_t *)PAGE8;
	*flashPtr = 0;                 //for flash virgin test
	flashPtr++;
	*flashPtr = ~(gVid & 0x0000FFFF);
	flashPtr++;
	while ((FLASH->SR & FLASH_SR_BSY) != 0) ;
	FLASH->SR |= FLASH_SR_EOP;
	if (FLASH->SR & FLASH_SR_OPERR)  return 0;

	*flashPtr = ~(gPid & 0x0000FFFF);
	flashPtr++;
	*flashPtr = ~gChannel;
	flashPtr++;
	while ((FLASH->SR & FLASH_SR_BSY) != 0) ;
	FLASH->SR |= FLASH_SR_EOP;
	if (FLASH->SR & FLASH_SR_OPERR)  return 0;

	*flashPtr = ~gdDmxMax;
	flashPtr++;
	*flashPtr = ~gdFadeTime;	
	flashPtr++;
	while ((FLASH->SR & FLASH_SR_BSY) != 0) ;
	FLASH->SR |= FLASH_SR_EOP;
	if (FLASH->SR & FLASH_SR_OPERR) return 0;
//this place provides a more safe storage of Vid/Pid
	*flashPtr = ~(gVid & 0x0000FFFF);
	flashPtr++;
	*flashPtr = ~(gPid & 0x0000FFFF);
	flashPtr++;
	while ((FLASH->SR & FLASH_SR_BSY) != 0) ;
	FLASH->SR |= FLASH_SR_EOP;
	if (FLASH->SR & FLASH_SR_OPERR)  return 0;

//terminate
    FLASH->CR &= ~FLASH_CR_PG;
//lock flash
    FLASH->CR |= FLASH_CR_LOCK;
    msDelay(10);
    return  1;
}


void popDmxFlash(uint32_t preset)  // ***************************************
// preset==0:all channels black. (1<=preset<=MAXPRESET) load preset from Flash
// *************************************************************************
{
uint32_t k, n,page, offset;
uint32_t * flashPtr;
uint32_t temp;

	if (preset == 0)  {
		for (k=0; k < DMX_BUFLEN;k++)  gdFINALMEM[k] = 0;//"dark" scene
	}
	else  {
		preset--;
		page = preset >> 2;								//0,1,2,...20
		offset = (preset - (page<<2)) * 512;			//0,512,1024,1536
		flashPtr = (uint32_t *)(PAGE9 + (2048 * page) + offset);
		for (k=0; k < 512;k+=4)  {
			n = k>>2;
			temp = ~(*flashPtr);
			gdFINALMEM[k] = (temp >> 24) & 0x000000FF;
			gdFINALMEM[k+1] = (temp >> 16) & 0x000000FF;
			gdFINALMEM[k+2] = (temp >> 8) & 0x000000FF;
			gdFINALMEM[k+3] = temp & 0x000000FF;
			flashPtr++;
		}
	}
	for (k=0;k < DMX_BUFLEN;k++)  {
		if (gdDMXBUF[k] == gdFINALMEM[k]) continue;
		if (gdFadeTime == 0)  gdDMXBUF[k] = gdFINALMEM[k];
		else startFader(k);
	}
}


uint32_t pushDmxFlash(uint32_t preset)   //**********************************
// **************************************************************************
{
uint32_t k, n, page, offset;
uint32_t * flashPtr;
uint32_t tempBuf[512];

	if (preset == 0)  {
		if (pushFlashPage8(1) == 0)  {
			sendUsbString(gcErrorStr);
			return 0;
		}
		return 1;				//preset#0 is all channels dark=0
	}
	preset--;

	page = preset >> 2;						//0,1,2,...15
	offset = (preset- (page<<2)) * 128;		//0,128,256,384 ->4 DMX presets in 1 page

	flashPtr = (uint32_t *)(PAGE9 + (page*2048));
//512 bytes must be located into 2k page and 2k page restored back to Flash
	for (k=0; k<512;k++)  {
		tempBuf[k] = ~(*flashPtr);
		flashPtr++;
	}
	for (n=0;n < 512;n+=4)  {
		k = n>>2;
		tempBuf[offset+k] = (gdDMXBUF[n]<<24) | (gdDMXBUF[n+1]<<16) |
										(gdDMXBUF[n+2]<<8) | gdDMXBUF[n+3] ;
	}
//unlock flash
#define KEY1    0x45670123      //mandatory unlock key defined by STM
#define KEY2    0xCDEF89AB

    while ((FLASH->SR &FLASH_SR_BSY) !=0); //wait until no operation is going on
    if ((FLASH->CR & FLASH_CR_LOCK) != 0) {
		FLASH->KEYR = KEY1;
		FLASH->KEYR = KEY2;
    }

//feed watchdog
    IWDG->KR = 0x0000AAAA;          //timeout ca. 1-2 seconds

//erase Page9 + page
    while ((FLASH->SR & FLASH_SR_BSY) != 0) ;  //wait until flash is not busy
    FLASH->CR = (((9 + page)<<3)|FLASH_CR_EOPIE|FLASH_CR_ERRIE|FLASH_CR_PER);
                                //EOPIE is set formally to get SR_EOP set
    FLASH->CR |= FLASH_CR_STRT;
    while ((FLASH->SR & FLASH_SR_BSY) != 0) ;
    FLASH->SR |= FLASH_SR_EOP;
    if (FLASH->SR & FLASH_SR_OPERR)  return 0;

    FLASH->CR &= ~FLASH_CR_PER;
//this microcontroller allows exclusively doubleword-wide write in flash
	FLASH->CR |= FLASH_CR_PG;
	flashPtr = (uint32_t *)(PAGE9 + 0x800 * page);
	for (k = 0; k < 512; k += 2)  {
		*flashPtr = ~(tempBuf[k]);
		flashPtr++;
		*flashPtr = ~(tempBuf[k+1]);
		flashPtr++;
//burn
		while ((FLASH->SR & FLASH_SR_BSY) != 0) ;
		FLASH->SR |= FLASH_SR_EOP;
		if (FLASH->SR & FLASH_SR_OPERR)  return 0;
	}
//terminate
    FLASH->CR &= ~FLASH_CR_PG;

//lock flash
    FLASH->CR |= FLASH_CR_LOCK;
    msDelay(10);
    return 1;
}

//###########################################################################

#ifdef RS232_DEBUG
void initUsart2(void) // ****************************************************
// USART2-TxD is PortA,2
// **************************************************************************
{
    RCC->APB1ENR1 |=  RCC_APB1ENR1_USART2EN ;   // USART1 clock enable

//set PA2 in alternate function mode
    GPIOA->MODER &= ~GPIO_MODER_MODE2_0 ;
    GPIOA->MODER |= GPIO_MODER_MODE2_1;
//USART1 is AF7 on PortB
    GPIOA->AFR[0] &= 0xFFFFF0FF;        //PA2 is TX
    GPIOA->AFR[0] |= 0x00000700;

    USART2->BRR = 0x01A1;	 //default 115200Baud: 416.667 at 48 MHz APB1 Clock

    USART2->CR2 = USART_CR2_TXINV;	//default:1 Stopbit, TX level inverted
    USART2->CR3 = 0;				//default:no Handshake
    USART2->CR1 = USART_CR1_UE | USART_CR1_TE; //8Bit, no Parity, no interrupt
}

void sendDBGB(uint8_t data) // *************************************************
//send 1 raw byte
// **************************************************************************
{
    while(!(USART2->ISR&USART_ISR_TXE));
    USART2->TDR = data;
}


void sendX8(uint8_t data)  // ********************************************
// **************************************************************************
{
uint8_t temp;

    temp= (data&0xF0) >> 4;
    temp = temp+0x30;
    if (temp>0x39) temp+= 7;
    sendDBGB(temp);
    temp= data&0x0F;
    temp = temp+0x30;
    if (temp>0x39) temp+= 7;
    sendDBGB(temp);
    sendDBGB(0x20);
}


void sendNibble(uint8_t data) // ********************************************
// **************************************************************************

{
uint8_t asciiNibble;

    asciiNibble = data+0x30;
    if (asciiNibble>0x39) asciiNibble+= 7;
    sendDBGB(asciiNibble);
}


void sendX32(uint32_t data) // **********************************************
// **************************************************************************
{
uint8_t nibble;

    nibble = (uint8_t) ((data>>28) & 0x0000000F);
    sendNibble(nibble);
    nibble = (uint8_t) ((data>>24) & 0x0000000F);
    sendNibble(nibble);
    nibble = (uint8_t) ((data>>20) & 0x0000000F);
    sendNibble(nibble);
    nibble = (uint8_t) ((data>>16) & 0x0000000F);
    sendNibble(nibble);
    nibble = (uint8_t) ((data>>12) & 0x0000000F);
    sendNibble(nibble);
    nibble = (uint8_t) ((data>>8) & 0x0000000F);
    sendNibble(nibble);
    nibble = (uint8_t) ((data>>4) & 0x0000000F);
    sendNibble(nibble);
    nibble = (uint8_t) (data & 0x0000000F);
    sendNibble(nibble);
    sendDBGB(0x20);
}
#endif

void sendUsbString(const uint8_t * str) // ***********************************
// ***************************************************************************
{
 uint8_t byte;

    while  ((byte = *str++) != 0)  sendUsbByte(byte);
}


void sendUsbX8(uint8_t data)  // ********************************************
// **************************************************************************
{
uint8_t temp;

    temp= (data&0xF0) >> 4;
    temp = temp+0x30;
    if (temp>0x39) temp+= 7;
    sendUsbByte(temp);
    temp= data&0x0F;
    temp = temp+0x30;
    if (temp>0x39) temp+= 7;
    sendUsbByte(temp);
    sendUsbByte(0x20);
}

void sendUsbN8(uint8_t data) // *********************************************
// send 1 byte in hex format (without leading 0x), WITHOUT TRAILING SPACE
// **************************************************************************
{
    uint8_t temp;

    temp= (data&0xF0) >> 4;
    temp = temp+0x30;
    if (temp>0x39) temp+= 7;
    sendUsbByte(temp);
    temp= data&0x0F;
    temp = temp+0x30;
    if (temp>0x39) temp+= 7;
    sendUsbByte(temp);
}


void sendUsbNibble(uint8_t data) // ********************************************
// **************************************************************************

{
uint8_t asciiNibble;

    asciiNibble = data+0x30;
    if (asciiNibble>0x39) asciiNibble+= 7;
    sendUsbByte(asciiNibble);
}


void sendUsbX16(uint16_t data) // *******************************************
// **************************************************************************
{
uint8_t nibble;

    nibble = (uint8_t) ((data>>12) & 0x000F);
    sendUsbNibble(nibble);
    nibble = (uint8_t) ((data>>8) & 0x000F);
    sendUsbNibble(nibble);
    nibble = (uint8_t) ((data>>4) & 0x000F);
    sendUsbNibble(nibble);
    nibble = (uint8_t) (data & 0x000F);
    sendUsbNibble(nibble);
    sendUsbByte(0x20);
}


void sendUsbX32(uint32_t data) // **********************************************
// **************************************************************************
{
uint8_t nibble;

    nibble = (uint8_t) ((data>>28) & 0x0000000F);
    sendUsbNibble(nibble);
    nibble = (uint8_t) ((data>>24) & 0x0000000F);
    sendUsbNibble(nibble);
    nibble = (uint8_t) ((data>>20) & 0x0000000F);
    sendUsbNibble(nibble);
    nibble = (uint8_t) ((data>>16) & 0x0000000F);
    sendUsbNibble(nibble);
    nibble = (uint8_t) ((data>>12) & 0x0000000F);
    sendUsbNibble(nibble);
    nibble = (uint8_t) ((data>>8) & 0x0000000F);
    sendUsbNibble(nibble);
    nibble = (uint8_t) ((data>>4) & 0x0000000F);
    sendUsbNibble(nibble);
    nibble = (uint8_t) (data & 0x0000000F);
    sendUsbNibble(nibble);
    sendUsbByte(0x20);
}


void sendUsbN32(uint32_t data) // **********************************************
//send signed decimal number string.
//Leading zeroes are suppressed, no trailing SPACE
//up to  +/- 2.147.483.647 Billions
// **************************************************************************
{
uint32_t quotient;

    if (data >= 0x80000000) {
        sendUsbByte('-');
        data = ~data + 1 ;
    }
    if (data < 10) goto ONES;
    if (data < 100) goto TENS;
    if (data < 1000) goto HUNDREDS;
    if (data < 10000) goto THOUSANDS;
    if (data < 100000) goto TH10;
    if (data < 1000000) goto TH100;
    if (data < 10000000) goto MILLIONS;
    if (data < 100000000) goto MIL10;
    if (data < 1000000000) goto MIL100;
    quotient = data/1000000000;
    data = data - quotient*1000000000;
    sendUsbByte((uint8_t)quotient+ 0x30);
MIL100:
    quotient = data/100000000;
    data = data - quotient*100000000;
    sendUsbByte((uint8_t)quotient + 0x30);
MIL10:
    quotient = data/10000000;
    data = data - quotient*10000000;
    sendUsbByte((uint8_t)quotient + 0x30);
MILLIONS:
    quotient = data/1000000;
    data = data - quotient*1000000;
    sendUsbByte((uint8_t)quotient + 0x30);
TH100:
    quotient = data/100000;
    data = data - quotient*100000;
    sendUsbByte((uint8_t)quotient+ 0x30);
TH10:
    quotient = data/10000;
    data = data - quotient*10000;
    sendUsbByte((uint8_t)quotient + 0x30);
THOUSANDS:
    quotient = data/1000;
    data = data - quotient*1000;
    sendUsbByte((uint8_t)quotient + 0x30);
HUNDREDS:
    quotient = data/100;
    data = data - quotient*100;
    sendUsbByte((uint8_t)quotient+ 0x30);
TENS:
    quotient = data/10;
    data = data - quotient*10;
    sendUsbByte((uint8_t)quotient + 0x30);
ONES:
    sendUsbByte((uint8_t)data + 0x30) ;
}


void sendUsbS32(uint32_t data) // **********************************************
//send signed decimal number string.
//Leading zeroes are suppressed, no trailing SPACE
//up to  +/- 2.147.483.647 Billions
// **************************************************************************
{
    sendUsbN32(data);
    sendUsbByte(0x20);
}


void sendUsbS16(uint16_t data) // ********************************************
//send signed decimal number string. Leading zeroes are suppressed
//up to  +/- 32767
// **************************************************************************
{
uint16_t quotient;

    if (data >= 0x8000) {
        sendUsbByte('-');
        data = ~data + 1 ;
    }
    if (data < 10) goto ONES;
    if (data < 100) goto TENS;
    if (data < 1000) goto HUNDREDS;
    if (data < 10000) goto THOUSANDS;
    quotient = data/10000;
    data = data - quotient*10000;
    sendUsbByte((uint8_t)quotient + 0x30);
THOUSANDS:
    quotient = data/1000;
    data = data - quotient*1000;
    sendUsbByte((uint8_t)quotient + 0x30);
HUNDREDS:
    quotient = data/100;
    data = data - quotient*100;
    sendUsbByte((uint8_t)quotient+ 0x30);
TENS:
    quotient = data/10;
    data = data - quotient*10;
    sendUsbByte((uint8_t)quotient + 0x30);
ONES:
    sendUsbByte((uint8_t)data + 0x30) ;
//    sendUsbByte(0x20);
}

// ##########################################################################
// Interrupt Handler

void SysTick_Handler(void)  // ***********************************************
// ***************************************************************************
{
	if ((gSYSCOUNT > 0)&&(gSYSCOUNT < 0x8000)) gSYSCOUNT--; //controls servSystem()
	else gSYSCOUNT = 0;             //double safety against underflow
	if (gLEDCOUNT > 0) gLEDCOUNT--;
	if (gDELAYCOUNT > 0) gDELAYCOUNT--;
	if (gTIMEOUT > 0) gTIMEOUT--;   //to be used in any subroutine
}

// ##########################################################################

